//===------------------------------------------------------------*- C++ -*-===//
//
// TEMU: The Terma Emulator
// (c) Terma 2021
// Authors: Daria Vorotnikova <davo (at) terma.com>
//
//===----------------------------------------------------------------------===//

#ifndef TEMU_BUS_PCIe_H
#define TEMU_BUS_PCIe_H
#include <assert.h>
#include <stddef.h>
#include <stdint.h>

#include "temu-c/Bus/PCIDefines.h"
#include "temu-c/Memory/Memory.h"
#include "temu-c/Support/Logging.h"
#include "temu-c/Support/Objsys.h"

#ifdef __cplusplus
extern "C" {
#endif
/**
 * temu_PCIeMessageTypes:
 * PCI Express Messages
 * name = code
 */
typedef enum {
  tePMT_AssertIntA = 32,
  tePMT_AssertIntB = 33,
  tePMT_AssertIntC = 34,
  tePMT_AssertIntD = 35,
  tePMT_DeassertIntA = 36,
  tePMT_DeassertIntB = 37,
  tePMT_DeassertIntC = 38,
  tePMT_DeassertIntD = 39,
  tePMT_PmActiveStateNak = 20,
  tePMT_PmPme = 24,
  tePMT_PmeTurnOff = 25,
  tePMT_PmToAck = 27,
  tePMT_ErrCor = 48,
  tePMT_ErrNonFatal = 49,
  tePMT_ErrFatal = 51,
  tePMT_Unlock = 0,
  tePMT_SetSlotPowerLimit = 80,
  tePMT_VendorDefinedType0 = 126,
  tePMT_VendorDefinedType1 = 127,
  tePMT_AttentionIndicatorOn = 65,
  tePMT_AttentionIndicatorBlink = 67,
  tePMT_AttentionIndicatorOff = 64,
  tePMT_PowerIndicatorOn = 69,
  tePMT_PowerIndicatorBlink = 71,
  tePMT_PowerIndicatorOff = 68,
  tePMT_AttentionButtonPressed = 72
} temu_PCIeMessageTypes;

/**
 * temu_PCIeConfigType0:
 * Fields that exists only in config type 0
 */
typedef struct {
  uint32_t BAR[6];
  uint32_t CardbusCISPointer;
  uint32_t SubsystemVendorID;
  uint32_t MaxLatMinGntIntPinLine;
} temu_PCIeConfigType0;
/**
 * temu_PCIeConfigType1:
 * Fields that exists only in config type 1
 */
typedef struct {
  uint32_t BAR0;
  uint32_t LatencyTimer2SubBusNumBusNum2PrimBusNum;
  uint32_t SecondaryStatusIOLimitBase;
  uint32_t MemoryLimitBase;
  uint32_t PrefetchableMemoryLimitBase;
  uint32_t PrefetchableBaseUpper;
  uint32_t PrefetchableLimitUpper;
  uint32_t UpperIOLimitIOBase;
  uint32_t BridgeCntrIntPinLine;
} temu_PCIeConfigType1;

typedef struct {
  // Power Mgmt Capabilities |Next Pointer (0x4C)| Power Mgmt Capability ID
  uint32_t PowerMgmtPointerCapabilities;
  // Data | .. | Power Management Status & Control
  uint32_t DataPowerMgmtStatusControl;
  // PCI Express Capabilities | Next Pointer | PCI Express Capability ID
  uint32_t PCIePointerCapabilities;
  uint32_t DeviceCapabilities;
  uint32_t DeviceStatusControl;
  uint32_t LinkCapabilities;
  uint32_t LinkStatusControl;
  uint32_t SlotCapabilities;  // RC mode only
  uint32_t SlotStatusControl; // RC mode only
  uint32_t RootControl;       // RC mode only
  uint32_t RootStatus;
  // MSI Message Control | Next Pointer (NULL)| MSI Message Capability ID
  uint32_t MsgControlPointerMsgCapabilityID;
  uint32_t MessageAddress;
  uint32_t UpperMessageAddress;
  uint32_t MessageData;
} temu_PCIeDeviceSpecificConfigSpace;

typedef struct {
  uint32_t LTSSMStateStatusReg;
  uint32_t ControllerCoreClockRatioReg;
  uint32_t PowerManagementTimerReg;
  uint32_t PMETimeOutReg;
  uint32_t SubsystemVendorIDUpdateReg;
  uint32_t ConfigurationReadyReg;
  uint32_t FlowControlUpdateTimeoutReg;
  uint32_t SecondaryStatusIrqMaskReg;
} temu_PCIeControllerInternalCSRs;
/**
 * temu_PCIExpressConfig:
 * PCI Express configuration space registers
 */
typedef struct {
  // PCI Compatible Configuration Header
  uint32_t DeviceVendorID;
  uint32_t StatusCommand;
  uint32_t ClassCodeRevID;
  uint32_t BISTHeaderTypeLatencyTimerCacheLineSize;
  uint32_t CapPointer;
  uint32_t ExpansionROMBaseAddress;
  temu_PCIeConfigType0 ConfigType0;
  temu_PCIeConfigType1 ConfigType1;

  // PCI Express Extended Configuration Registers
  uint16_t AdvancedErrorReportingCapIdReg;
  uint32_t UncorrectableErrorStatusReg;
  uint32_t UncorrectableErrorMaskReg;
  uint32_t UncorrectableErrorSeverityReg;
  uint32_t CorrectableErrorStatusReg;
  uint32_t CorrectableErrorMaskReg;
  uint32_t AdvancedErrorCapabilitiesControlReg;
  uint32_t HeaderLogReg[4];
  uint32_t RootErrorCommandReg;
  uint32_t RootErrorStatusReg;
  uint32_t CorrectableErrorSourceIdReg;

  temu_PCIeDeviceSpecificConfigSpace
      PCIeDevConfig; // 0x40 to 0xFF via capabilities
  temu_PCIeControllerInternalCSRs PCIeController;
} temu_PCIExpressConfig;

typedef struct {
  void (*mapPciIo)(temu_Object *Obj, unsigned Device, unsigned Bar,
                   uint64_t Addr, uint64_t Len);
  void (*unmapPciIo)(temu_Object *Obj, unsigned Device);
  void (*mapPciMem)(temu_Object *Obj, unsigned Device, unsigned Bar,
                    uint64_t Addr, uint64_t Len);
  void (*unmapPciMem)(temu_Object *Obj, unsigned Device);
} temu_PCIExpressBusIface;
TEMU_IFACE_REFERENCE_TYPE(temu_PCIExpressBus)
#define TEMU_PCIe_BUS_IFACE_TYPE "temu::PCIExpressBusIface"

typedef struct {
  void (*raiseReset)(temu_Object *);
  void (*lowerReset)(temu_Object *);
  void (*connect)(temu_Object *, temu_Object *);
  void (*disconnect)(temu_Object *, temu_Object *);
  void (*sendMessage)(void *, temu_PCIeMessageTypes msgType, uint8_t *payload);
  void (*setUpstreamBridge)(temu_Object *);
} temu_PCIExpressBridgeIface;
TEMU_IFACE_REFERENCE_TYPE(temu_PCIExpressBridge)
#define TEMU_PCIe_BRIDGE_IFACE_TYPE "temu::PCIExpressBridgeIface"

typedef struct {
  temu_PCIExpressConfig *(*getPciExpressConfig)(temu_Object *);
  // Writes to config space forwarded here
  void (*writeConfig)(temu_Object *, uint32_t offset, uint32_t value);
  uint32_t (*readConfig)(temu_Object *, uint32_t offset);
  uint64_t (*getPciExpressExpansionROMSize)(temu_Object *);
  temu_MemAccessIface *(*getBARInterface)(temu_Object *, uint32_t index,
                                          uint8_t type);
} temu_PCIExpressDeviceIface;
TEMU_IFACE_REFERENCE_TYPE(temu_PCIExpressDevice)
#define TEMU_PCIe_DEVICE_IFACE_TYPE "temu::PCIExpressDeviceIface"

typedef struct temu_PCIExpressBus temu_PCIExpressBus;
typedef struct temu_PCIExpressBridge temu_PCIExpressBridge;

typedef struct {
  temu_Object Super;
  temu_PCIExpressConfig ExpressConf;
  temu_PCIExpressBus *PrimaryBus;
} temu_PCIExpressDevice;

struct temu_PCIExpressBus {
  temu_Object Super;
  uint8_t BusId;

  temu_PCIExpressBridge *ParentDev;
  temu_PCIExpressDeviceIfaceRefArray ChildrenDevs;
};

struct temu_PCIExpressBridge {
  temu_PCIExpressDevice Super;
  temu_PCIExpressBus *SecondaryBus;
  uint8_t PrimBusId;
  uint8_t SecBusId;
  uint8_t SubBusId;
};

static inline temu_PCIExpressBus *
temu_pciGetRootBus(temu_PCIExpressDevice *device)
{
  temu_PCIExpressBus *bus = device->PrimaryBus;

  while (bus->ParentDev != NULL) {
    device = &bus->ParentDev->Super;
    bus = device->PrimaryBus;
  }
  return bus;
}

static inline void
temu_pcieSetPrimaryBusId(temu_PCIExpressBridge *C, uint8_t Val)
{
  C->PrimBusId = Val;
}
static inline void
temu_pcieSetSecondaryBusId(temu_PCIExpressBridge *C, uint8_t Val)
{
  C->SecBusId = Val;
}
static inline void
temu_pcieSetSubordinaryBusId(temu_PCIExpressBridge *C, uint8_t Val)
{
  C->SubBusId = Val;
}

static inline void
temu_pcieSetBusId(temu_PCIExpressBus *C, uint8_t Val)
{
  C->BusId = Val;
}
static inline void
temu_pcieSetDeviceId(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->DeviceVendorID = ((uint32_t)Val << 16) | (C->DeviceVendorID & 0xffff);
}

static inline void
temu_pcieSetVendorId(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->DeviceVendorID = (uint32_t)Val | (C->DeviceVendorID & 0xffff0000);
}

static inline void
temu_pcieSetStatus(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->StatusCommand = ((uint32_t)Val << 16) | (C->StatusCommand & 0xffff);
}
static inline uint16_t
temu_pcieReadStatus(temu_PCIExpressConfig *C)
{
  return (uint16_t)((C->StatusCommand & 0xffff0000) >> 16);
}

static inline void
temu_pcieSetClassCode(temu_PCIExpressConfig *C, uint32_t Val)
{
  C->ClassCodeRevID &= 0x000000ff;
  C->ClassCodeRevID |= Val << 8;
}

static inline void
temu_pcieSetRevId(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ClassCodeRevID &= 0xffffff00;
  C->ClassCodeRevID |= Val;
}

static inline void
temu_pcieSetBist(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->BISTHeaderTypeLatencyTimerCacheLineSize =
      ((uint32_t)Val << 24) |
      (C->BISTHeaderTypeLatencyTimerCacheLineSize & 0x00ffffff);
}

static inline void
temu_pcieSetHeaderType(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->BISTHeaderTypeLatencyTimerCacheLineSize =
      ((uint32_t)Val << 16) |
      (C->BISTHeaderTypeLatencyTimerCacheLineSize & 0xff00ffff);
}

static inline void
temu_pcieSetLatencyTimer(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->BISTHeaderTypeLatencyTimerCacheLineSize =
      ((uint32_t)Val << 8) |
      (C->BISTHeaderTypeLatencyTimerCacheLineSize & 0xffff00ff);
}

static inline void
temu_pcieSetCacheLineSize(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->BISTHeaderTypeLatencyTimerCacheLineSize =
      ((uint32_t)Val << 0) |
      (C->BISTHeaderTypeLatencyTimerCacheLineSize & 0xffffff00);
}

static inline void
temu_pcieSetInterruptPin0(temu_PCIExpressConfig *C, uint8_t Val)
{
  assert(Val < 0x05);
  C->ConfigType0.MaxLatMinGntIntPinLine =
      ((uint32_t)Val << 8) |
      (C->ConfigType0.MaxLatMinGntIntPinLine & 0xffff00ff);
}

static inline void
temu_pcieSetSubsystemId(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType0.SubsystemVendorID =
      (uint32_t)Val << 16 | (C->ConfigType0.SubsystemVendorID & 0x0000ffff);
}

static inline void
temu_pcieSetSubsystemVendorId(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType0.SubsystemVendorID =
      (uint32_t)Val << 0 | (C->ConfigType0.SubsystemVendorID & 0xffff0000);
}
static inline void
temu_pcieSetPrimaryBusNumber(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum =
      (uint32_t)Val |
      (C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum & 0xffffff00);
}
static inline void
temu_pcieSetSecondaryBusNumber(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum =
      ((uint32_t)Val << 8) |
      (C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum & 0xffff00ff);
}
static inline void
temu_pcieSetSubordinateBusNumber(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum =
      ((uint32_t)Val << 16) |
      (C->ConfigType1.LatencyTimer2SubBusNumBusNum2PrimBusNum & 0xff00ffff);
}
static inline void
temu_pcieSetIOBase(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.SecondaryStatusIOLimitBase =
      ((uint32_t)Val) |
      (C->ConfigType1.SecondaryStatusIOLimitBase & 0xffffff00);
}
static inline void
temu_pcieSetIOLimit(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.SecondaryStatusIOLimitBase =
      ((uint32_t)Val << 8) |
      (C->ConfigType1.SecondaryStatusIOLimitBase & 0xffff00ff);
}
static inline void
temu_pcieSetSecondaryStatus(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.SecondaryStatusIOLimitBase =
      ((uint32_t)Val << 16) |
      (C->ConfigType1.SecondaryStatusIOLimitBase & 0x0000ffff);
}
static inline void
temu_pcieSetMemoryBase(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.MemoryLimitBase =
      ((uint32_t)Val) | (C->ConfigType1.MemoryLimitBase & 0xffff0000);
}
static inline void
temu_pcieSetMemoryLimit(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.MemoryLimitBase =
      ((uint32_t)Val << 16) | (C->ConfigType1.MemoryLimitBase & 0x0000ffff);
}
static inline void
temu_pcieSetPrefetchableMemoryBase(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.PrefetchableMemoryLimitBase =
      ((uint32_t)Val) |
      (C->ConfigType1.PrefetchableMemoryLimitBase & 0xffff0000);
}
static inline void
temu_pcieSetPrefetchableMemoryLimit(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.PrefetchableMemoryLimitBase =
      ((uint32_t)Val << 16) |
      (C->ConfigType1.PrefetchableMemoryLimitBase & 0x0000ffff);
}
static inline void
temu_pcieSetInterruptLine(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.BridgeCntrIntPinLine =
      ((uint32_t)Val) | (C->ConfigType1.BridgeCntrIntPinLine & 0xffffff00);
}
static inline void
temu_pcieSetInterruptPin1(temu_PCIExpressConfig *C, uint8_t Val)
{
  C->ConfigType1.BridgeCntrIntPinLine =
      ((uint32_t)Val << 8) | (C->ConfigType1.BridgeCntrIntPinLine & 0xffff00ff);
}
static inline void
temu_pcieSetBridgeControl(temu_PCIExpressConfig *C, uint16_t Val)
{
  C->ConfigType1.BridgeCntrIntPinLine =
      ((uint32_t)Val << 16) |
      (C->ConfigType1.BridgeCntrIntPinLine & 0x0000ffff);
}

static inline uint16_t
temu_pcieReadBridgeControl(temu_PCIExpressConfig *C)
{
  return (uint16_t)((C->ConfigType1.BridgeCntrIntPinLine & 0xffff0000) >> 16);
}

static inline uint32_t
temu_pcieReadConfigurationReadyReg(temu_PCIExpressConfig *C)
{
  return C->PCIeController.ConfigurationReadyReg;
}

static inline void
temu_pcieSetConfigurationReadyReg(temu_PCIExpressConfig *C, uint32_t Val)
{
  C->PCIeController.ConfigurationReadyReg = Val;
}

#ifdef __cplusplus
}
#endif
#endif // !TEMU_BUS_PCIe_H
