//===-- temu-c/Scheduler.h - TEMU Scheduler API -----------------*- C++ -*-===//
//
// TEMU: The Terma Emulator
// (c) Terma 2023
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//
#ifndef TEMU_SCHEDULER_H
#define TEMU_SCHEDULER_H
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct temu_TimeSource temu_TimeSource;

typedef enum temu_ExecutionState {
  teES_Stopped,
  teES_Running,
  teES_Stepping,
} temu_ExecutionState;


/*!
 * Run the current scheduler for a fixed number of steps
 * 
 * \param steps Time in steps
 */
void temu_runForSteps(int64_t steps);


/*!
 * Run the current scheduler for a fixed time
 * 
 * \param secs Time in simulated seconds
 */
void temu_runForTime(double secs);

/*!
 * Run the current scheduler forever
 */
void temu_run();

/*!
 * Step a processor in the current scheduler
 *
 * Note that, the other processors may run if the step exceeds the quanta.
 * The step function currently runs processors in sequence,
 * it is not running processors in parallel.
 * This means that stepping is always deterministic.
 * 
 * \param cpu The processor to step
 * \param steps The number of steps or instructions to execute
 */
void temu_step(temu_TimeSource *cpu, int steps);

/*!
 * Stop the current scheduler
 *
 * The function signals the scheduler to stop,
 * and then waits until the scheduler is in the stopped state.
 */
void temu_stop();
/*!
 * Stop the current scheduler
 * 
 * The function returns immediately, before the scheduler has stopped.
 * The function is suitable for use in e.g. signal handlers.
 */
void temu_asyncStop();

/*!
 * Get the current execution status of TEMU
 *
 * \result Current execution status
 */

temu_ExecutionState temu_getExecutionState();

#ifdef __cplusplus
} // extern c
#endif

#endif /* ! TEMU_REGISTER_H */
